
/* Handle Automatic Features */

#include "gnutella.h"

#include "interface.h"

#include "auto.h"

#include "search.h"

#include <sys/stat.h>

// todo notes: re-scan shared at certain interval

gboolean auto_or(char *, char *);

gboolean auto_enabled = FALSE; // global, flag to show we are working
guint32 auto_save_search_reissue_timeout; // global, re set this value at stop or quit
GSList *sl_auto_record = NULL; // global, pointer to all the list records

guint32 auto_timer = 0;
gint auto_net_stable = 0; // seconds the net has been stable (nodes connected)
gint auto_picked_row = 0; // the list row we picked to use
gint auto_work_row = 0; // the list row we ar working on now
gint auto_search_row = 0; // the search result list row we are on now
gint indicate_cnt = 0; // used for the little gui indicator that we are working
guint32 search_timer = 999999; // count time between searches
guint32 search_timeout = 0; // count time we haven't gotten a new result back
guint32 queue_count = 0; // how many files have been queued from this search
gboolean auto_stopped = FALSE; // flag if we are working or not


// Variables set by the auto config file

gboolean auto_on_at_start = FALSE; // global, Run at program startup?
guint32 auto_none_found = 10; // Wait xx minutes if no files are found, set in gui too
guint32 auto_queued_stop = 200; // This many files in the download queue, kill the search and wait
guint32 auto_queued_start = 20; // This many files in the download queue, start searching again
guint32 auto_search_time = 600; // How many seconds must pass before allowing another search
guint32 auto_search_timeout = 180; // Seconds to wait after last results come back before cancel
guint32 auto_flush_timeout = 3600; // Max seconds file will sit in queue, get rid of old data


// define the keywords to read from the automation config file

enum
{
	kw_none_found = 0, kw_queued_stop, kw_queued_start, kw_search_time, kw_search_timeout,
	kw_flush_timeout, kw_list_minsize, kw_list_not, kw_list_search, kw_on_at_start, kw_end
};

gchar *kw[] =
{
	"none_found",		// kw_none_found
	"queued_stop",		// kw_queued_stop
	"queued_start",	// kw_queued_start
	"search_time",		// kw_search_time
	"search_timeout",	// kw_search_timeout
	"flush_timeout",	// kw_flush_timeout
	"list_minsize",	// kw_list_minsize
	"list_not",			// kw_list_not
	"list_search",		// kw_list_search
	"on_at_start",		// kw_on_at_start
	NULL
};

gchar auto_tmp[4096];
gchar buf_list_minsize[4096];
gchar buf_list_not[4096];
gchar buf_list_search[4096];



// We check the file against minsize and the "not" list, this is
// where you could do more checking, you have access to all file data
// Be fast, we check up to 10 files at a time in the loop

gboolean auto_file_select(struct record *rc, struct auto_record *r, gboolean D)
{
	struct stat st;
	guint32 tmp_size;
	gboolean flag1 = TRUE;


	// Check file against minsize

	if (r->minsize < 0) { // negative numbers indicate do a "max size"
		if ((tmp_size = abs(r->minsize)) > 0) { // make it positve and check it
			tmp_size = tmp_size * 1000; // K Bytes
			if (rc->size > tmp_size) flag1 = FALSE; // yes, file is too big
			}
		}
	if (r->minsize > 0) { // check to see if it passes min size
		tmp_size = r->minsize * 1000; // K Bytes
		if (rc->size < tmp_size) flag1 = FALSE; // yes, file is not big enough
		}

	if (!flag1) {
		if (D) printf("Not Queued (size) '%s'\n",rc->name);
		return FALSE;
		}


	// Check file against "not" string

	if (auto_or(rc->name,r->notstr)) {
		if (D) printf("Not Queued '%s'\n",rc->name);
		return FALSE;
		}


	// Check picked file in "done" dir, is it there, if so is it a larger size?

	g_snprintf(auto_tmp, sizeof(auto_tmp), "%s/%s", move_file_path, rc->name);
	if (stat(auto_tmp, &st) != -1) { // does the picked file exist?
		if (rc->size < st.st_size) { // is it smaller?
			if (D) printf("Not Queued (dup & size) '%s'\n",rc->name);
			return FALSE;
			}
		if (rc->size == st.st_size) { // exact size?
			if (D) printf("Not Queued (dup) '%s'\n",rc->name);
			return FALSE;
			}
		}

	return TRUE; // file passes, queue it
}



// All the automatic logic, called once every second if auto_enabled
// Here we simulate a very patient, non greedy human brain working all night.
// Try to be as nice as possible to the network at all times.
// If you modify this code you will have to spend the hours it takes to
// test it properly. Comment everything, do it the easy-reading way.
// We depend on the downloads module to filter dups and files we already have

// NOTE: This first version is very simple, more work and testing is needed.
//
// Pick up search term from top of list
// Send search out and wait for results
// Queue all the files that pass the "not" filter up to max
// Get next search term from list, do it all again
//

void auto_brain()
{
	gboolean D = FALSE; // debug printout, turn this on during testing of this routine

	struct auto_record *r;
	struct search *sch;
	struct results_set *rs;
	struct record *rc;
	GSList *l;
	gint i;
	gchar tbuf[260];
	time_t tloc;
	gchar indicate_buf[50], indicate_char;
	gchar *indicate_str = "|/-\\"; // 4 characters

	auto_timer++;
	search_timer++;
	search_timeout++;

	indicate_char = indicate_str[indicate_cnt++];
	if (indicate_cnt > 3) indicate_cnt = 0;
	if (auto_timer > 86400)
		g_snprintf(indicate_buf, sizeof(indicate_buf), "       %ud %uh %c",
				auto_timer / 86400, (auto_timer % 86400) / 3600, indicate_char);
	else if (auto_timer > 3600)
				g_snprintf(indicate_buf, sizeof(indicate_buf), "       %uh %um %c",
						auto_timer / 3600, (auto_timer % 3600) / 60, indicate_char);
	else if (auto_timer > 60)
				g_snprintf(indicate_buf, sizeof(indicate_buf), "        %um %us",
						auto_timer / 60, auto_timer % 60);
	else g_snprintf(indicate_buf, sizeof(indicate_buf), "          %us", auto_timer);
	gtk_label_set(GTK_LABEL(label_info_2), indicate_buf); // working indicator

	if (on_the_net()) { // check if host connections are stable
		if (++auto_net_stable > 20) { // 20 seconds is a good number
			auto_net_stable = 30; // keep it within limits
			}
		}
	else {
		auto_net_stable = 0; // start count over
		return;
		}

	if (auto_net_stable < 20) return; // go no further if net is not stable

	if (auto_stopped) {
		if (GTK_CLIST(clist_download_queue)->rows <= auto_queued_start) { // almost out of files, start again
			auto_stopped = FALSE;
			if (D) printf("Auto queued start\n");
			}
		else return;
		}
	else {
		if (GTK_CLIST(clist_download_queue)->rows >= auto_queued_stop) { // enough files queued for now
			while (searches) search_close_current(); // close all searches, next time use new one from list
			auto_stopped = TRUE;
			if (D) printf("Auto queued stop\n");
			gtk_label_set(GTK_LABEL(label_info_3), "        (idle)");
			return;
			}
		}


	if (!searches) { // start a new search if one is not going
		if (auto_picked_row >= GTK_CLIST(clist_auto)->rows) auto_picked_row = 0; // cycle back to top of list
		r = (struct auto_record *) gtk_clist_get_row_data(GTK_CLIST(clist_auto), auto_picked_row);
		if (r) {
			if (search_timer > auto_search_time) {
				new_search(0, r->search); // enter a new search
				auto_work_row = auto_picked_row;
				auto_picked_row++; // set up for next time
				auto_search_row = 0; // start at the top of the search list
				search_timer = 0;
				search_timeout = 0;
				queue_count = 0;
				if (D) {
					time(&tloc); // do a time stamp
					strftime(tbuf,sizeof(tbuf),"%d %H:%M:%S",localtime(&tloc));
					printf("\n----- %s SEARCH '%s'\n",tbuf,r->search);
					}
				gtk_label_set(GTK_LABEL(label_info_3), "     (searching)");
				g_snprintf(indicate_buf, sizeof(indicate_buf), "Searching");
				gtk_clist_set_text(GTK_CLIST(clist_auto), auto_work_row, 3, indicate_buf);
				}
			}
		}

		// Here's where it all happens

	else {
		l = searches;
		sch = (struct search *) l->data;
		if (sch) { // we should always have a search record if "searches" is valid
			i = 0;

			while(1) {
					// get a search result from the list
				rc = (struct record *) gtk_clist_get_row_data(GTK_CLIST(sch->clist), auto_search_row);
				if (rc) { // no new results yet ?
					search_timeout = 0; // reset the timeout, we have a result
					rs = rc->results_set; // contains the filename and other data
						// get the "not" string and minsize from our auto list
					r = (struct auto_record *) gtk_clist_get_row_data(GTK_CLIST(clist_auto), auto_work_row);
					if (r) { // check r just in case, but it should be ok
						if (auto_file_select(rc, r, D)) { // should we queue this file?
							download_new(rc->name, rc->name, rc->size, rc->index, rs->ip, rs->port, rs->guid); // queue it
							if (D) printf("Queue '%s'\n",rc->name);
							g_snprintf(indicate_buf, sizeof(indicate_buf), "   (queued %u)", queue_count);
							gtk_label_set(GTK_LABEL(label_info_3), indicate_buf); // status
							g_snprintf(indicate_buf, sizeof(indicate_buf), "Queued %u", queue_count);
							gtk_clist_set_text(GTK_CLIST(clist_auto), auto_work_row, 3, indicate_buf);
							queue_count++;
							}
						if (++auto_search_row >= GTK_CLIST(sch->clist)->rows) break; // next row, till end
						if (++i >= 10) break; // only do up to 10 filenames this pass
						}
					}
				else { // no new results yet
					if (search_timeout > auto_search_timeout) { // no new results for quite some time now
						while (searches) search_close_current(); // close all searches, next time use new one from list
						}
					break;
					}
				} // close the while loop
			}
		} // close else

}




// called after program starts
void auto_init()
{
	if (auto_on_at_start) auto_startup();
}


// startup the automatic feature, when they hit the button
void auto_startup()
{
	while (searches) search_close_current(); // close all searches

	auto_enabled = TRUE;
	auto_timer = 0;
	auto_net_stable = 0;
	auto_picked_row = 0;
	auto_work_row = 0;
	auto_search_row = 0;
	indicate_cnt = 0;
	search_timer = 999999;
	search_timeout = 0;
	queue_count = 0;
	auto_stopped = FALSE;

	auto_save_search_reissue_timeout = search_reissue_timeout;
	search_reissue_timeout = 0; // we control search reissue now
	gtk_widget_set_sensitive(button_auto_start, FALSE);
	gtk_widget_set_sensitive(button_auto_stop, TRUE);
	gtk_label_set(GTK_LABEL(label_info_1), "AUTOMATION ON");

		// disable entry, passive, filter & stop button on search screen
	gtk_widget_set_sensitive(button_search_passive, FALSE);
	gtk_widget_set_sensitive(button_search_filter, FALSE);
	gtk_widget_set_sensitive(button_search_stop, FALSE);
	gtk_widget_set_sensitive (entry_search, FALSE);
}


// Stop the automatic mode, when they hit the button
void auto_shutdown()
{
	while (searches) search_close_current(); // close all searches
	auto_enabled = FALSE;
	search_reissue_timeout = auto_save_search_reissue_timeout;
	if (sl_auto_record) gtk_widget_set_sensitive(button_auto_start, TRUE);
	gtk_widget_set_sensitive(button_auto_stop, FALSE);
	gtk_label_set(GTK_LABEL(label_info_1), "");
	gtk_label_set(GTK_LABEL(label_info_2), "");
	gtk_label_set(GTK_LABEL(label_info_3), "");

		// enter, passive & filter button on search screen
	gtk_widget_set_sensitive(button_search_passive, TRUE);
//	gtk_widget_set_sensitive(button_search_filter, TRUE); // disabled for now
	gtk_widget_set_sensitive (entry_search, TRUE);
}


void auto_remove_entry(struct auto_record *r, gint row)
{
	gtk_clist_remove(GTK_CLIST(clist_auto), row);

	sl_auto_record = g_slist_remove(sl_auto_record, (gpointer) r);

	g_free(r->search);
	g_free(r->notstr);
	g_free(r);
}


void str_dejunk(char *str)
{

  int c = 0, x = 0;

  c = strlen(str);

  for (;x<c;x++)
    if (str[x] < 34 || str[x] > 125) str[x] = ' ';

  return;
}


// New entry

void auto_new_entry(gchar *search, gchar *notstr, gint minsize)
{
	struct auto_record *r;
	gchar *titles[4];
	gint row;
	gchar astmp_1[40]; // used for the minsize string

	r = (struct auto_record *) g_malloc0(sizeof(struct auto_record));

	r->status = 0;
	r->search = g_strdup(search);
	r->notstr = g_strdup(notstr);
	r->minsize = minsize;
	r->time = 0;
	r->info = NULL;

	str_dejunk(r->search);
	g_strdown(r->search);
	str_dejunk(r->notstr);
	g_strdown(r->notstr);


	// add it to our list of records
	sl_auto_record = g_slist_prepend(sl_auto_record, (gpointer) r);

	titles[0] = r->search;
	titles[1] = r->notstr;
	g_snprintf(astmp_1, sizeof(astmp_1), "%dKB", r->minsize);
	titles[2] = astmp_1;
	titles[3] = NULL;

	row = gtk_clist_append(GTK_CLIST(clist_auto), titles);
	gtk_clist_set_row_data(GTK_CLIST(clist_auto), row, (gpointer) r);

	gtk_widget_set_sensitive(button_auto_start, TRUE);

}


// Do a full OR match of the filename against the "not" string (already lowercase)

gboolean auto_or(char *s_filename, char *notstr)
{
  char *a;
  char *q2;
  char *s2;
  char *loc;

  // Copy notstr and filename so strtok doesn't trash them
  q2 = g_strdup(notstr);
  s2 = g_strdup(s_filename);
  g_strdown(s2);

	// break the notstr string into words ("tokens"). For example,
	// if the notstr is "john smith .jpg" the tokens will be "john",
	// "smith" and "jpg" tabs are handled as spaces

  a = strtok(q2, " \t"); // get first token
  if (a == 0) {
    g_free(q2);
    g_free(s2);
    return FALSE;
  }

	// Check each token to see if it's in the string

  while (a) {
    loc = strstr(s2, a); // is there a match?
    if (loc) { // found it
		g_free(q2);
		g_free(s2);
	 	return TRUE; // it's in there
		}
    a = strtok(0, " \t"); // go to next token
  }

	g_free(q2);
	g_free(s2);
	return FALSE; // if we got this far then it's not in there
}


// used to set the variables and load the list

void auto_set_param(guint32 keyword, gchar *value)
{
	gint32 i = atol(value);

	switch (keyword)
	{
		case kw_none_found: { if (i > 1 && i < 1048576) auto_none_found = i; return; }
		case kw_queued_stop: { if (i > 1 && i < 1048576) auto_queued_stop = i; return; }
		case kw_queued_start: { if (i > 1 && i < 1048576) auto_queued_start = i; return; }
		case kw_search_time: { if (i > 300 && i < 1048576) auto_search_time = i; return; }
		case kw_search_timeout: { if (i > 60 && i < 1048576) auto_search_timeout = i; return; }
		case kw_flush_timeout: { if (i > 1800 && i < 1048576) auto_flush_timeout = i; return; }
		case kw_list_minsize: { strncpy(buf_list_minsize, value, sizeof(buf_list_minsize)); return; }
		case kw_list_not: { strncpy(buf_list_not, value, sizeof(buf_list_not)); return; }
		case kw_on_at_start: { auto_on_at_start = i ? TRUE : FALSE; return; }

		// special case, enter all into the clist, including previous values
		case kw_list_search: {
			strncpy(buf_list_search, value, sizeof(buf_list_search)); // limit the size
			buf_list_minsize[sizeof(buf_list_minsize)-1] = 0;
			buf_list_not[sizeof(buf_list_not)-1] = 0; // strncpy doesn't terminate if at max
			buf_list_search[sizeof(buf_list_search)-1] = 0;
			auto_new_entry(buf_list_search, buf_list_not, atol(buf_list_minsize));
			buf_list_minsize[0] = 0; // clear these
			buf_list_not[0] = 0;
			return;
			}
	}
}


// read in automation settings from file

void auto_read_from_file(gchar *path, gboolean quiet)
{
	FILE *f;
	gchar *s, *k, *v;
	guint32 i, n = 0;

	static gchar *err = "Bad line %u in NAPS-auto.txt file, ignored\n";
	buf_list_minsize[0] = 0; // in case someone forgets these
	buf_list_not[0] = 0;

	f = fopen(path, "r");

	if (!f)
	{
		if (!quiet) g_warning("Unable to open file %s (%s)\n", path, g_strerror(errno));
		return;
	}

	while (fgets(auto_tmp, sizeof(auto_tmp), f)) // fgets terminates string if at max size
	{
		n++;
		s = auto_tmp;
		while (*s && (*s == ' ' || *s == '\t')) s++;
		if (!((*s >= 'A' && *s <= 'Z') || (*s >= 'a' && *s <= 'z'))) continue;
		k = s;
		while (*s =='_' || (*s >= 'A' && *s <= 'Z') || (*s >= 'a' && *s <= 'z') || (*s >= '0' && *s <= '9')) s++;
		if (*s != '=' && *s != ' ' && *s != '\t') { fprintf(stderr, err, n); continue; }
		v = s;
		while (*s == ' ' || *s == '\t') s++;
		if (*s != '=') { fprintf(stderr, err, n); continue; }
		*v = 0; s++;
		while (*s == ' ' || *s == '\t') s++;
		if (*s == '"')
		{
			v = ++s;
			while (*s && *s != '\n' && *s != '"') s++;
			if (!*s || *s == '\n') { fprintf(stderr, err, n); continue; }
		}
		else { v = s; while (*s && *s != '\n' && *s != ' ' && *s != '\t') s++; }
		*s = 0;

		for (i = 0; i < kw_end; i++) if (!g_strcasecmp(k, kw[i])) { auto_set_param(i, v); break; }

		if (i >= kw_end) fprintf(stderr, "'NAPS-auto.txt' file, line %u: unknown keyword '%s', ignored\n", n, k);
	}

	fclose(f);

}


// write out the automation settings

void auto_write_to_file(gchar *path)
{

	FILE *f;
	struct auto_record *r;
	gint row = 0;
	f = fopen(path, "w");

	if (!f) {
		g_warning("Unable to open output file %s (%s)\n", path, g_strerror(errno));
		return;
		}

	fprintf(f, "\n# This is the NapShare automation configuration file.\n");
	fprintf(f, "# You may edit it if you're careful. See the README file for more info.\n");
	fprintf(f, "# Edit it when the program is not running, it is written over after quit.\n\n");

	fprintf(f, "# Wait xx minutes if no files are found, set via gui\n"
			"%s = %u\n\n", kw[kw_none_found], auto_none_found);
	fprintf(f, "# This many files in the download queue, kill the search and wait\n"
			"%s = %u\n\n", kw[kw_queued_stop], auto_queued_stop);
	fprintf(f, "# This many files in the download queue, start searching again\n"
			"%s = %u\n\n", kw[kw_queued_start], auto_queued_start);
	fprintf(f, "# How many seconds must pass before allowing another search (>300)\n"
			"%s = %u\n\n", kw[kw_search_time], auto_search_time);
	fprintf(f, "# Seconds to wait after last results come back before cancel (>60)\n"
			"%s = %u\n\n", kw[kw_search_timeout], auto_search_timeout);
	fprintf(f, "# Max seconds file will sit in queue, get rid of old data (>1800)\n"
			"%s = %u\n\n", kw[kw_flush_timeout], auto_flush_timeout);
	fprintf(f, "# Start automation when program starts?\n"
			"%s = %u\n\n", kw[kw_on_at_start], auto_on_at_start);

	fprintf(f, "\n# The following values are entered into the automation list at each\n");
	fprintf(f, "# occurrence of the value for \"search\", set \"min size\" and \"not\"\n");
	fprintf(f, "# first. 4K limit on each string.\n\n");

	while (row < GTK_CLIST(clist_auto)->rows) { // write out the list
		r = (struct auto_record *) gtk_clist_get_row_data(GTK_CLIST(clist_auto), row);
		fprintf(f, "%s = \"%d\"\n", kw[kw_list_minsize], r->minsize);
		fprintf(f, "%s = \"%s\"\n", kw[kw_list_not], r->notstr);
		fprintf(f, "%s = \"%s\"\n\n", kw[kw_list_search], r->search);
		row++;
		}

	fprintf(f, "\n\n");

	fclose(f);
}


